package cn.ibizlab.core.extensions.service;

import cn.ibizlab.core.lite.extensions.service.DbEntityService;
import cn.ibizlab.core.rule.domain.ExecResult;
import cn.ibizlab.core.rule.service.impl.ExecResultServiceImpl;
import com.alibaba.druid.pool.DruidDataSource;
import com.baomidou.dynamic.datasource.DynamicRoutingDataSource;
import com.baomidou.dynamic.datasource.toolkit.DynamicDataSourceContextHolder;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Lazy;
import org.springframework.context.annotation.Primary;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.ObjectUtils;
import org.springframework.util.StringUtils;

import javax.sql.DataSource;
import java.util.*;

/**
 * 实体[执行结果] 自定义服务对象
 */
@Slf4j
@Primary
@Service("ExecResultExService")
public class ExecResultExService extends ExecResultServiceImpl {

    @Autowired
    private DstDataSourceExService dstDataSourceService;

    @Autowired
    private DbEntityService dbEntityService;

    @Autowired
    @Lazy
    private DynamicRoutingDataSource dynamicRoutingDataSource;

    private Map<String, List<ExecResult>> resultsSaveAsync = new HashMap<>();
    private Object objResultSaveAsyncLock = new Object();
    private Timer timerSaveAsync = null;

    private Map<String, List<ExecResult>> resultsDeleteAsync = new HashMap<>();
    private Object objResultDeleteAsyncLock = new Object();
    private Timer timerDeleteAsync = null;

    @Override
    protected Class currentModelClass() {
        return com.baomidou.mybatisplus.core.toolkit.ReflectionKit.getSuperClassGenericType(this.getClass().getSuperclass(), 1);
    }

    /**
     * [SaveAsync:SaveAsync] 行为扩展
     *
     * @param et
     * @return
     */
    @Override
    @Transactional
    public ExecResult saveAsync(ExecResult et) {
        synchronized (objResultSaveAsyncLock) {
            addExecResult(resultsSaveAsync, et);
        }
        //创建一个定时器
        if (timerSaveAsync == null) {
            timerSaveAsync = new Timer();
            //schedule方法是执行时间定时任务的方法
            timerSaveAsync.schedule(new TimerTask() {

                //run方法就是具体需要定时执行的任务
                @Override
                public void run() {
                    try {
                        synchronized (objResultSaveAsyncLock) {
                            if (resultsSaveAsync.size() > 0) {
                                for (Map.Entry<String, List<ExecResult>> entry : resultsSaveAsync.entrySet()) {
                                    String tabName = "";
                                    String dsName = entry.getKey();
                                    List<ExecResult> execResults = entry.getValue();
                                    List<ExecResult> tempExecResults = new ArrayList<>();
                                    if (execResults.size() > 0) {
                                        dstDataSourceService.initDataSource(dsName);
                                        DynamicDataSourceContextHolder.push(dsName);
                                        for (ExecResult execResult : execResults) {
                                            tempExecResults.add(execResult);
                                            Object objTabName = execResult.get(RuleEngineExService.Setting_ResultTableName);
                                            if (StringUtils.isEmpty(tabName) && !ObjectUtils.isEmpty(objTabName)) {
                                                tabName = String.valueOf(objTabName);
                                            }
                                            if (tempExecResults.size() >= 500) {
                                                saveResultBatch(tempExecResults, dsName, tabName);
                                            }
                                        }
                                        if (tempExecResults.size() > 0) {
                                            saveResultBatch(tempExecResults, dsName, tabName);
                                        }
                                        resultsSaveAsync.remove(dsName);
                                    }
                                }
                            }
                        }
                    } catch (Exception ex) {
                        log.error(ex.getMessage());
                    } finally {
                        DynamicDataSourceContextHolder.poll();
                    }

                }
            }, 1000, 10000);
        }
        return et;
    }

    /**
     * [DeleteAsync:DeleteAsync] 行为扩展
     *
     * @param et
     * @return
     */
    @Override
    @Transactional
    public ExecResult deleteAsync(ExecResult et) {
        synchronized (objResultDeleteAsyncLock) {
            addExecResult(resultsDeleteAsync, et);
        }
        if (timerDeleteAsync == null) {
            timerDeleteAsync = new Timer();
            //schedule方法是执行时间定时任务的方法
            timerDeleteAsync.schedule(new TimerTask() {

                //run方法就是具体需要定时执行的任务
                @Override
                public void run() {
                    try {
                        synchronized (objResultDeleteAsyncLock) {
                            if (resultsDeleteAsync.size() > 0) {
                                for (Map.Entry<String, List<ExecResult>> entry : resultsDeleteAsync.entrySet()) {
                                    String tabName = "";
                                    String dsName = entry.getKey();
                                    List<ExecResult> execResults = entry.getValue();
                                    List<ExecResult> tempExecResults = new ArrayList<>();
                                    if (execResults.size() > 0) {
                                        dstDataSourceService.initDataSource(dsName);
                                        DynamicDataSourceContextHolder.push(dsName);
                                        for (ExecResult execResult : execResults) {
                                            tempExecResults.add(execResult);
                                            Object objTabName = execResult.get(RuleEngineExService.Setting_ResultTableName);
                                            if (StringUtils.isEmpty(tabName) && !ObjectUtils.isEmpty(objTabName)) {
                                                tabName = String.valueOf(objTabName);
                                            }
                                            if (tempExecResults.size() >= 500) {
                                                dbEntityService.clearResultBatch(tempExecResults, tabName);
                                            }
                                        }
                                        if (tempExecResults.size() > 0) {
                                            dbEntityService.clearResultBatch(tempExecResults, tabName);
                                        }
                                        resultsDeleteAsync.remove(dsName);
                                    }
                                }
                            }
                        }
                    } catch (Exception ex) {
                        log.error(ex.getMessage());
                    } finally {
                        DynamicDataSourceContextHolder.poll();
                    }

                }
            }, 1000, 10000);
        }
        return et;
    }

    /**
     * 将规则执行结果添加到待保存集合中
     * @param execResultMap
     * @param execResult
     */
    private void addExecResult(Map<String, List<ExecResult>> execResultMap, ExecResult execResult) {
        Object dsName = execResult.get(RuleEngineExService.Setting_ResultDataSource);
        Object tabName = execResult.get(RuleEngineExService.Setting_ResultTableName);
        if (ObjectUtils.isEmpty(tabName)) {
            execResult.set(RuleEngineExService.Setting_ResultTableName, RuleEngineExService.Default_ResultTableName);
        }
        if (ObjectUtils.isEmpty(dsName)) {
            dsName = RuleEngineExService.Default_ResultDataSource;
        }
        String dataSource = String.valueOf(dsName);
        if (execResultMap.containsKey(dataSource)) {
            execResultMap.get(dataSource).add(execResult);
        } else {
            execResultMap.put(dataSource, new ArrayList() {{
                add(execResult);
            }});
        }
    }

    /**
     * 根据数据源类型保存结果
     * @param args
     * @param
     */
    private void saveResultBatch(List<ExecResult> args, String dsName, String tabName) {
        Map<String, DataSource> dynamicDSMap = dynamicRoutingDataSource.getCurrentDataSources();
        if (!ObjectUtils.isEmpty(dynamicDSMap) && dynamicDSMap.containsKey(dsName)) {
            DataSource ds = dynamicDSMap.get(dsName);
            if (ds instanceof DruidDataSource) {
                DruidDataSource druidDataSource = (DruidDataSource) dynamicDSMap.get(dsName);
                if ("oracle".equals(druidDataSource.getDbType())) {
                    dbEntityService.saveResultBatchByOracle(args, tabName);
                    args.clear();
                    return;
                }
            }
        }
        dbEntityService.saveResultBatch(args, tabName);
        args.clear();
    }
}

